# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
# from April 2020 in GitHub Dictation-toolbox Unimacro
#
# Mouseless Browsing extension from Rudolf Noe (mode = "mlb") is discarded
#
# only clickbyvoice, which is present in all browsers that are chromium compatible
# chrome, safari, brave, edge (Microsoft Edge) etc.
#
# 
#
# written by: Quintijn Hoogenboom (QH software, training & advies)
#
# the lists {n1-9}, {n1-20} are constructed in internal grammar functions
#
# the lists {pagecommands} and {tabcommands} in the inifile (edit chrome hah)
#
#pylint:disable=C0209

"""
This commands grammar allows clicking by voice

It is a global grammar, that is activated as soon as one of the chromium browsers is
in the foreground 

"""
import natlink
from unimacro import natlinkutilsbj as natbj
from dtactions.unimacro import unimacroutils
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doKeystroke  as keystroke

# use extension Click by Voice
visiblePause = 0.4

language = unimacroutils.getLanguage()

ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):

    try:
        numberGram = natbj.numberGrammarTill999[language]
    except KeyError:
        print('take number grammar from "enx"')
        numberGram = natbj.numberGrammarTill999['enx']
        
    if language == "nld":
        name = 'Click by voice'
    else:
        name = 'Click by voice'

    gramSpec = """
<shownumbers> exported = ((show) (numbers) [{additionalonoroff}]+) | ((numbers) {additionalonoroff}+) ;
<hidenumbers> exported = (hide) (numbers);
<picknumber> exported = (pick) <integer> [{navigateoptions}];

# is already in _tasks grammar:
# <navigatetabs> exported = ((next|previous) tab) [{n1-20} | {tabcommands}]|
#                            (tab (back|forward) [{n1-20} | {tabcommands}]);
# <numberedtabs> exported = tab {n1-n20} [{tabcommands}];
# <tabactions> exported = tab {tabcommands};


<navigatepages> exported = ((next|previous|{pagecommands}) page)|
                            (page (back|forward) [{n1-20}]) |
                            page {pagecommands} |
                            (next|previous) page {pagecommands};
                            
#and the numbers grammar (0,...,999 or chain of digits):                             
"""+numberGram
        
    def initialize(self):
        self.prevHandle = -1
        self.ActiveHndle = None
        self.load(self.gramSpec)

    def gotBegin(self,moduleInfo):
        if not language:
            return
        if self.checkForChanges:
            self.prevHandle = None

        winHandle = moduleInfo[2]
        if not winHandle:
            print(f'no window handle in {self.name}')
            return
        if self.prevHandle == winHandle:
            return
        self.prevHandle = winHandle
        progInfo = unimacroutils.getProgInfo(moduleInfo)
        # print('progInfo: %s'% repr(progInfo))
        prog = progInfo.prog
        chromiumBrowsers = {'chromium', 'chrome', 'msedge', 'safari', 'brave'}
        if prog in chromiumBrowsers:
            if progInfo.toporchild == 'child':
                print(f'in child window, of a clickbyvoice program {prog}')
            if self.checkForChanges:
                print(f'_clickbyvoice ({self.name}, prog: {prog}, checking the inifile')
                self.checkInifile()
            self.switchOnOrOff(window=winHandle)
            if not self.ActiveHndle == winHandle:
                print(f'activate _clickbyvoice, {prog}, hndle: {winHandle}')
                self.activateAll(window=winHandle)
                self.ActiveHndle = winHandle
        else:
            if self.isActive():
                print("deactivate _clickbyvoice")
                self.deactivateAll()
                self.ActiveHndle = False
                

    def gotResultsInit(self,words,fullResults):
        """at start of actions"""
        self.number = ''
        self.navOption = ''   # eg left or right (s or o)hallo
        self.hadPick = False

    def gotResults_picknumber(self, words, fullResults):
        """start the collecting of the wanted number
        """
        if self.hasCommon(words[0], 'pick'):
            self.waitForNumber('number')
            self.hadPick = True
        else:
            self.navOption = self.getFromInifile(words[-1], 'navigateoptions')
            
    def gotResults_shownumbers(self, words, fullResults):
        """show the numbers, with additional options

        """
        # print 'showhidenumbers, words: %s'% words
        showNumbers = ":+"  # fresh start, just in case
        additionalOptions = False
        while 1:
            additional = self.getFromInifile(words[-1], 'additionalonoroff', noWarning=1)
            if additional is None:
                break
            if additional == '-':
                print(f'{self.name}: hide the numbers')
                self.gotResults_hidenumbers(words, fullResults)
                return
            words.pop() # remove last word of list.
            if additional in showNumbers:
                pass
            else:
                showNumbers += additional 
                additionalOptions = True
                
        if additionalOptions:
            print(f'{self.name}: showNumbers command: {showNumbers}, set as new default for the current session.')
            # set new chosen string:
            # self.setInInifile("general", "show numbers", showNumbers)

        self.showNumbers = showNumbers
        self.getInputcontrol()
        self.doOption(showNumbers)
        print('clickbyvoice, before finishInputControl')
        self.finishInputControl()
        print('clickbyvoice, after finishInputControl')


    def gotResults_hidenumbers(self, words, fullResults):
        """hide the numbers

        """
        self.getInputcontrol()
        self.doOption(self.hideNumbers)
        self.finishInputControl()

    def gotResults_tabactions(self,words,fullResults):
        """do an actions to the current tab (doc)"""
        # print(f'tabactions words: {words}')
        command = self.getFromInifile(words, 'tabcommands')
            
        if command:
            action(command)

    def gotResults_numberedtabs(self,words,fullResults):
        """go to a numbered tab (doc) and do an optional action"""
        print(f'numberedtabs: {words}')
        command = self.getFromInifile(words, 'tabcommands')

        counts = self.getNumbersFromSpoken(words)
        if not counts:
            print(f'_clickbyvoice, numberedtabs, no valid tab number found: {words}')
            return
            
        if command:
            action(command)

    def gotResults_navigatetabs(self,words,fullResults):
        """go to next or previous tab(s) (documents) and refresh possibly"""
        print(f'navigate tabs: {words}')
        direction = None
        command = self.getFromInifile(words, 'tabcommands',noWarning=1)
        
        if self.hasCommon(words, ['next', 'verder', 'volgende', 'vooruit', 'forward']):
            direction = 'tab'
        elif self.hasCommon(words, ['previous', 'terug', 'vorige', 'back']):
            direction = 'shift+tab'
        else:
            print(f'no direction found in command: {words}')
        
        counts = self.getNumbersFromSpoken(words)
        if counts:
            count = counts[0]
        else:
            count = 1
##        print 'tabs:     direction: %s, count: |%s|, command: |%s|'% (direction, counlinker balkt, command)

        if direction:        
            while count > 0:
                count -= 1
                keys = '{ctrl+' + direction + '}'
                keystroke(keys)
                unimacroutils.Wait(0.5) #0.3 seem too short for going back tabs in chrome
            
        if command:
            action(command)

    def gotResults_navigatepages(self,words,fullResults):
        """go to next or previous page(s) and refresh possibly"""
##        print 'navigate pages: %s'% words
        ## not active at the moment, possibly reactivate...
        direction = None
        command = self.getFromInifile(words, 'pagecommands',noWarning=1)
        
        if self.hasCommon(words, ['next', 'verder', 'volgende', 'vooruit', 'forward']):
            direction= 'right'
        elif self.hasCommon(words, ['previous', 'terug', 'vorige', 'back']):
            direction= 'left'
        else:
            print(f'no direction found in command: {words}')
        
        counts = self.getNumbersFromSpoken(words)
        if counts:
            count = counts[0]
        else:
            count = 1
##        print 'PAGES:     dir: %s, count: |%s|, command: |%s|'% (dir, counlinker balkt, command)

        if direction:        
            while count > 0:
                count= count -1
                keystroke('{alt+%s}'%(direction))
                unimacroutils.Wait(0.5) #0.3 seem too short for going back pages in chrome
            
        if command:
            action(command)


    def gotResults(self,words,fullResults):
    
        # step 5, in got results collect the number:
        # only necessary in this grammar for collecting the choose command
        if not self.hadPick:
            return
        self.collectNumber()
        if not self.number:
            print('collected no number')
            return
        self.getInputcontrol()
        command = self.number
        commandparts = []
        if self.navOption:
            if not self.navOption.startswith(":"):
                command += ":"
            command += self.navOption
        if command.find(';') >= 0:
            print(f'command: {command}')
            commandparts = command.split(';')
            command = commandparts.pop(0)
            print(f'command: {command}, commandparts: {commandparts}')
        self.doOption(command)
        for additional in commandparts:
            unimacroutils.Wait(visiblePause)
            keystroke(additional)
        self.finishInputControl()
        
        
        
    def getInputcontrol(self):
        """get the Click by Voice input control"""
        keystroke("{shift+ctrl+space}")
        unimacroutils.Wait()   ## longer: unimacroutils.Wait(visiblePause)
        for i in range(10):
            progInfo = unimacroutils.getProgInfo()
            if progInfo.toporchild == 'child':
                if i:
                    print(f'found input window after {i} steps')
                break
            unimacroutils.Wait()
        else:
            print("_clickbyvoice failed to reach input window")
        # print("found input window of clickbyvoice")
        unimacroutils.visibleWait()
        
        
    def doOption(self, option):
        """after the inputcontrol is focussed, do the command"""
        keystroke(option)

    def finishInputControl(self):
        """press enter, after a little bit of waiting
        """
        unimacroutils.visibleWait()
        unimacroutils.visibleWait()
        keystroke("{enter}")
        
    def fillInstanceVariables(self):
        """fills the necessary instance variables

        """
        self.showNumbers = self.ini.get('general', 'show numbers') or ':+'
        if self.showNumbers.find(":") != 0:
            if self.showNumbers.find(":") == -1:
                self.showNumbers = ":" + self.showNumbers
            else:
                print(f'{self.name}, "+" sign missing in inifile, "general", "show numbers": "{self.showNumbers}", replace by default: ":+"')
                self.showNumbers = ":+"
        if self.showNumbers.find("+") != 1:
            print('{self.name}, "+" sign missing or in wrong position in inifile, "general", "show numbers": "{self.showNumbers}", replace by default: ":+"')
            self.showNumbers = ":+"
        # not in inifile:
        self.hideNumbers = ":-"



# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
if __name__ == "__main__":
    ## interactive use, for debugging:
    with natlink.natConnect():
        try:
            thisGrammar = ThisGrammar(inifile_stem="_clickbyvoice")
            # thisGrammar.startInifile()
            thisGrammar.initialize()
            print('clickbyvoice, before finishInputControl')
            thisGrammar.finishInputControl()
            print('clickbyvoice, after finishInputControl')



        finally:
            thisGrammar.unload()
elif __name__.find('.') == -1:
    # standard startup when Dragon starts:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()

    
def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None




"""Unimacro grammar to Dictate latex markup, as defined in an inifile

"""
# This file is/was part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
# _latex.py: Dictate latex markup
#
# written by: Frank Olaf Sem-Jacobsen
# March 2011
#
#pylint:disable = R0912, C0209, E1101
import natlink
from natlinkcore import nsformat 
from natlinkcore import natlinkstatus
from dtactions.unimacro import unimacroutils
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.sendkeys import sendkeys as keystroke
import unimacro.natlinkutilsbj as natbj

status = natlinkstatus.NatlinkStatus()
language = status.language
ICAlphabet = natbj.getICAlphabet(language=language)

# import re
# reBracedActions = re.compile(r"\{")

ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):

    name = "latex"
    gramSpec = """
<dgndictation> imported;
<structure> exported =  begin {floating} [<dgndictation>];
<reference> exported =  reference ({floating} | {label}) <dgndictation>;
<namereference> exported =  name reference ({floating} | {label}) <dgndictation>;
<label> exported =  label ({floating} | {label}) <dgndictation>;
<command> exported =  {commands} [with] [ (arguments [{arguments}]) | (that | ([this] line)) ] [(and | with) label];
<options> exported =  add option [{options}];
<replace> exported = replace {placeholders} [with <dgndictation>]; 
    """

    def initialize(self):
        if not language:
            print("no valid language in grammar "+__name__+" grammar not initialized")
            return

        self.load(self.gramSpec)
        self.switchOnOrOff()

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile()

    def gotResultsInit(self,words,fullResults):
        self.dictation = ''
        self.floating = ''
        self.label = 0
        self.reference = ''
        self.namereference = ''
        self.label_text=''
        self.Cmd = ''
        self.replace_text = 0

    def gotResults_replace(self, words, fullResults):
        place = self.getFromInifile(words, 'placeholders', noWarning=1)
        action('<<startsearch>>')
        keystroke(place)
        action('<<searchgo>>')
        action('<<cancel>>')
        if self.hasCommon(words, ['with']):
            self.replace_text = 1

    def gotResults_command(self, words, fullResults):
        initial=self.getFromInifile(words, 'commands', noWarning=1)
        pos = initial.find('{')
        print(pos)
        contents =''
        if self.hasCommon(words, ['arguments']):
            if pos > 0:
                Cmd = initial[:pos + 1]
            else:
                Cmd = initial+'{'
            args  = self.getFromInifile(words, 'arguments', noWarning=1)
            Cmd = Cmd+args + '}'
            print(Cmd)
        else:
            Cmd = initial
        if self.hasCommon(words, ['that']):
            contents = self.get_selection_that(line = 0)
        if self.hasCommon(words, ['line']):
            contents = self.get_selection_that(line = 1)
        contents = contents.strip()

        stringpaste(Cmd)
        if pos > 0  or self.hasCommon(words, ['arguments']):
            keystroke('{left}')
        if contents:
            if pos ==  - 1:
                stringpaste('{')
            stringpaste(contents)
            if pos == -1:
                stringpaste('}')
            else:
                keystroke('{right}')
        if self.hasCommon(words, ['label']):
            label =self.getFromInifile(words, 'label', noWarning=1)
            if label:
                ## workaround for keystrokes: {{}
                keystroke('{enter}')
                stringpaste(r'\label{%s}'% (self.makes_label(label, contents)))
                keystroke('{enter}')
                
    def gotResults_options(self, words, fullResults):
        selection = self.view_selection_current_line()
        if selection:
            pass
            # print(f'select_current_line: {selection}')
        else:
            print('no selection found')
        options = self.getFromInifile(words, 'options', noWarning=1)
        present = 1
        squared = selection.find(']')
        only = 1
        if squared  == -1:
            squared = selection.find('{')
            present = 0
        else:
            squared = squared +1
            start = selection.find ('[')
            if start!= squared -2:
                only = 0
        if squared ==  - 1:
            keystroke('{end}')
        else:
            keystroke('{home}')
            for _ in range(0, squared):
                keystroke('{right}')
            if present == 0:
                keystroke('[')
            if options:
                if present == 1:
                    keystroke('{left}')
                if only == 0:
                    keystroke (',')
                keystroke(options)
            if present == 0:
                keystroke(']')
            if not options:
                keystroke('{left}')
            

    def gotResults_reference(self, words, fullResults):
        self.reference = self.getFromInifile(words, 'floating', noWarning=1) or \
                         self.getFromInifile(words, 'label', noWarning = 1)
                         
    def gotResults_namereference(self, words, fullResults):
        self.namereference = self.getFromInifile(words, 'floating', noWarning=1) or \
                         self.getFromInifile(words, 'label', noWarning = 1)                  

    def gotResults_label(self, words, fullResults):
        self.label_text = self.getFromInifile(words, 'floating', noWarning=1) or \
                         self.getFromInifile(words, 'label', noWarning = 1)
                                                 
    def gotResults_structure(self, words, fullResults):
        self.floating=self.getFromInifile(words, 'floating', noWarning=1)
      
    
    def gotResults(self, words, fullResults):
        if self.floating:
            floatingString = r'\begin{%s}'% self.floating
            stringpaste(floatingString)
            keystroke('{enter}')
            if self.dictation:
                labelString = self.makes_label(self.floating, self.dictation)
                dictationString = r'\label{%s}'% labelString
                stringpaste(dictationString)
            keystroke ('{enter}')
            floatingString = '\\end{%s}'% self.floating
            stringpaste(floatingString)
            keystroke('{up}')
            print('floating: %s'% self.floating)
        if self.reference:
            stringpaste('\\ref{%s}' % (self.makes_label(self.reference, self.dictation)))
        if self.namereference:
            stringpaste('\\nameref{%s}' % (self.makes_label(self.namereference, self.dictation)))
        if self.label_text:
            stringpaste('\\label{%s}' % (self.makes_label(self.label_text, self.dictation)))
        if self.replace_text:
            stringpaste(self.dictation)
        

    def gotResults_dgndictation(self, words, fullResults):
        """do with nsformat functions"""
        # print('got dgndictation: %s'% words)
        self.dictation, dummy = nsformat.formatWords(words)  # state not needed in call
        # print('   result of nsformat:  %s'% repr(self.dictation))


    def get_selection_that(self, line = 0):
        unimacroutils.saveClipboard()

        if line:
            action('<<selectline>><<cut>>')
        else:
            action('<<cut>>')
        action('W')
        contents = natlink.getClipboard().strip().replace('\r', '')
        if len(contents) == 0:
            if line:
                print('_latex, empty line')
                return ""
            action('HW select that')
            action('<<cut>>')
            action('W')
            contents = natlink.getClipboard().strip().replace('\r', '')
            if len(contents) == 0:
                print('_latex, empty contents, no last dictate utterance available')
                
        unimacroutils.restoreClipboard()
        return contents

    def view_selection_current_line(self):
        unimacroutils.saveClipboard()
        keystroke('{ctrl+c}')
        action('W')
        contents = natlink.getClipboard()
        if len(contents) == 0:
            # print('no_space_by_existing selection')
            keystroke('{end}{shift+home}')
            keystroke('{ctrl+c}')
            action('W')
            contents = natlink.getClipboard()
        unimacroutils.restoreClipboard()
        return contents


    def makes_label(self, Type, Text):
        return Type + ':' + ''.join(Text.split()).lower()
  
def stringpaste(t):
    """paste via clipboard, to circumvent German keyboard issues
    """
    print(f'stringpaste: |{t}|')
    action(f'SCLIP "{t}"')
    
# standard stuff Joel (QH, Unimacro, python3):
try:
    thisGrammar
except NameError:
    thisGrammar = None

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None

if __name__ == "__main__":
    # here code to interactive run this module
    natlink.natConnect()
    try:
        thisGrammar = ThisGrammar(inifile_stem='_latex')
        thisGrammar.startInifile()
        thisGrammar.initialize()
        Words = ['add', 'option', 'draft']
        FR = {}
        thisGrammar.gotResults_options(Words, FR)
    finally:
        natlink.natDisconnect()
elif __name__.find('.') == -1:
    # called from the loader, when starting Dragon/Natlink:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()



"""input box, in use with the _folders grammar

"""
#pylint:disable=E1101
import wx


def InputBox(text, prompt="Unimacro Input", default=""):
    """this one kills Dragon"""
    _app = wx.App()
    dialog = wx.TextEntryDialog(None,
                text, prompt, default, style=wx.OK|wx.CANCEL)
    if dialog.ShowModal() == wx.ID_OK:
        result = dialog.GetValue()
    else:
        result = None
    dialog.Destroy()
    return result

if __name__ == "__main__":
    _result = InputBox("just testing")
    print(_result, repr(_result))
    


__version__ = "$Rev: 429 $ on $Date: 2011-05-31 16:21:03 +0200 (di, 31 mei 2011) $ by $Author: quintijn $"
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
# Excel navigate in a sheet
"""
"""


import natlink
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
import pprint
import types
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro import unimacroactions as actions
import win32com

# 
language = unimacroutils.getLanguage()
ancestor = natbj.DocstringGrammar
class ThisGrammar(ancestor):

    try:
        numberGram = natbj.numberGrammarTill999[language]
    except KeyError:
        print('take number grammar from "enx"')
        numberGram = natbj.numberGrammarTill999['enx']
        
    name = 'excel'
    iniIgnoreGrammarLists = ['sheets', 'books']
    gramSpec = """
<col> exported = column ({character}|{customcolumn}|back)[<excelcolumncommand>|<negativeaction>];
<date> exported = [<before>] date ({n1-31} | {n1-31}{month} | {n1-31}{month}{year})|
                        {datespecial};
<excelcolumncommand> = {excelcolumncommands};
<books> exported = (workbook) ({books}|show);
<sheets> exported = (sheet) ({sheets}|show);
<negative> exported = make number <negativeaction>;
<negativeaction> = negative | positive;
<before> = here;
#and the numbers grammar (0,...,999):                             
"""+numberGram
        
    def initialize(self):
        self.prevHandle = -1
        self.load(self.gramSpec)
        self.prevRows = []
        self.prevCols = []
        self.excel = None
        self.sheetsList = []
        self.booksList = []
        
    def gotBegin(self,moduleInfo):
        winHandle = moduleInfo[2]
        if self.prevHandle == winHandle:
            if not self.isActive():
                return
        else:
            self.prevHandle = winHandle
            if unimacroutils.matchModule('excel', modInfo=moduleInfo):
                #print 'activate firefox %s mode'% mode
                if self.checkForChanges:
                    print('excel (%s), checking the inifile'% self.name)
                    self.checkInifile()
                self.switchOnOrOff()
            elif self.isActive():
                #print 'deactivate Excel %s mode'% mode
                self.deactivateAll()
                if self.excel:
                    self.excel.disconnect()
                    self.excel = None
                return
        if self.isActive():
            # refreshes current position now as well:
            progInfo = unimacroutils.getProgInfo(moduleInfo)
            if self.excel is None:
                self.excel = actions.get_instance_from_progInfo(progInfo)
                if self.excel.app:
                    print('excel.app: %s'% self.excel.app)
                    print('Workbooks: %s'% self.excel.app.Workbooks.Count)
                else:
                    print('no instance for this Excel object available')
                    self.deactivateAll()
                    return
                    
            res = self.excel.checkForChanges(progInfo)
            #print 'res from changes: %s'% res
            if res:   
                self.refreshExcelSettings(res)
                self.printPositions('from gotBegin')

    #  Bij het initialiseren wordt de horizontale spatiejering uitgerekend.
    #    Er wordt van uitgegaan dat het venster niet te smal staat.
    def gotResultsInit(self,words,fullResults):
        """at start of actions"""
        # just in case a button was kept down (with show card commands)
        self.hadChoose = None
        self.chooseSpecification = ''
        self.number = ''
        self.hadRow = self.hadCol = None
        res = self.excel.checkForChanges()
        if res:
            self.refreshExcelSettings(res)
            self.printPositions('from gotResultsInit')

    def refreshExcelSettings(self, res):
        """here set settings of instance variables, when state has changed
        """
        if res <= 1:
            print('only position')
            return
        sheets = self.excel.getSheetsList()
        if not sheets:
            print('warning Excel: no sheets found in getSheetsList')
        if sheets != self.sheetsList:
            self.sheetsList = sheets
            self.sheetsDict = self.spokenforms.getDictOfMixedSpokenForms(sheets)
            self.sheetsList = list(self.sheetsDict.keys())
            self.setList('sheets', self.sheetsList)
            print('sheets: %s'% self.sheetsList)
        books = self.excel.getBooksList()
        if books != self.booksList:
            self.booksList = books
            self.booksDict = self.spokenforms.getDictOfMixedSpokenForms(books)
            self.booksList = list(self.booksDict.keys())
            self.setList('books', self.booksList)
            print('books: %s'% self.booksList)

    # several commands in Docstring format:
    def rule_backgroundcolor(self, words):
        """[here] background [color] {color}
        """
        app = self.excel.app
        if self.hasCommon(words[0], 'here'):
            action("MP 2,0,0;")
        colorCode = int(self.getFromInifile(words[-1], 'color'))
        if colorCode:
            app.ActiveCell.Interior.ColorIndex = colorCode
        else:
            app.ActiveCell.Interior.ColorIndex = None
    #                    
    #
    #        
    #def gotResults_row(self,words,fullResults):
    #    if self.hasCommon(words, 'back'):
    #        prevRow = self.popFromListIfDifferent(self.prevRows, self.currentRow)
    #        if prevRow:                
    #            self.gotoRow(prevRow)
    #        else:
    #            print 'no row to go back to'
    #        return
    #    self.hadRow = 1
    #    self.waitForNumber('number')


    def gotResults_books(self,words,fullResults):
        print('books: %s'% words)
        print(self.booksList)

    def gotResults_sheets(self,words,fullResults):
        print('sheets: %s'% words)
        print(self.sheetsList)
        self.excel.selectSheet(self.sheetsDict[words[1]])        

    def gotResults_negativeaction(self,words,fullResults):
        #print 'excel, make active cell negative'
        n = self.excel.app.ActiveCell.Value
        if type(n) == float:
            n = -n
            self.excel.app.ActiveCell.Value = n
        else:
            print('excel, not a float: %s (%s)'% (n, type(n)))

    def gotResults_col(self,words,fullResults):
        self.hadCol = 1
        if self.hasCommon(words[1], 'back'):
            prevCol = self.excel.getPreviousColumn()
            if prevCol:                
                self.gotoCol(prevCol)
            else:
                print('no column to go back to')
            return
        col = self.getCharacterFromSpoken(words[1])
        if col:
            self.gotoCol(col)
            return
        col = self.getFromInifile(words[1], 'customcolumn')
        if col:
            self.gotoCol(col)
        else:
            raise ValueError('excel, column: no valid column found: "%s" (all words: %s)'% (words[1], words))

    def gotResults_before(self,words,fullResults):
        if self.hasCommon(words, 'here'):
            if not self.doWaitForMouseToStop():
                print('excel, mouse does not stop, cancel command')
                return
            unimacroutils.buttonClick()

    def gotResults_date(self,words,fullResults):
        day = self.getNumberFromSpoken(words[1])
        if len(words) > 2:
            month = self.getFromInifile(words[2], 'month')
        else:
            month = 12
        keystroke('%s-%s{tab}'% (day, month))

    def gotoCol(self, col):
        """col should be a single letter
        """
        sheet = self.excel.sheet
        col = col.strip()
        if col == self.excel.currentColumn:
            print('column %s already selected'% col)
            return
        #print 'col: %s, curRow: %s'% (col, self.excel.currentRow)
        range = col + self.excel.currentRow
        sheet.Range(range).Select()
        cr = self.excel.savePosition()
        #print 'all current positions:'
        #pprint.pprint(self.excel.positions)

    def printPositions(self, comment=None):
        """print position info, for debugging purposes
        """
        if comment:
            print('----%s'% comment)
        print('colum info: %s'% self.excel.columns)
        print('row info: %s'% self.excel.rows)
        print('---------')
        

# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
thisGrammar = ThisGrammar()
if thisGrammar.gramSpec:
    thisGrammar.initialize()
else:
    thisGrammar = None
    
def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar: thisGrammar.unload()
    thisGrammar = None




# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
# _keystrokes.py 
#  written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  August 2002, revised September 2003, revised June 2011
"""grammar in which you can continuously dictate keystrokes

There are repeatable keys, like up, down, tab, and non-repeatable keys
like 'alpha, bravo, Home, Context Menu'.

Modes are introduced in this grammar so that in some programs special
keystrokes can be defined, that can be used in combination with all the
others continuously.

"""
#pylint:disable=R0904, R0913, R0912, C0209

import sys
import copy
from dtactions.unimacro import inivars
from dtactions.unimacro.unimacroactions import doKeystroke as keystroke
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro import unimacroutils
from unimacro import natlinkutilsbj as natbj
from natlinkcore import nsformat 
from natlinkcore import loader
import natlink
natlinkmain = loader.NatlinkMain()
language = unimacroutils.getLanguage()        

ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):

    iniIgnoreGrammarLists = ['count', 'effcount', 'character', 'punctuation']
    name = 'keystrokes'
    stdkeysrule = """<stdkeys> = <contextmenu> | <repkey> | <repkey><count> | <norepkey> |
                     <modifiers><repkey> | <modifiers><repkey><count> | <modifiers><norepkey>;"""
    allkeysrule = """<allkeys> = <lastkey> |
        ((<click>|<stdkeys>)+ | (<click>|<stdkeys>)+<lastkey>);"""
    
    lastkeyrule = "<lastkey> = dictate <dgndictation>;"  # only with allkeys!
    importedrule = "<dgndictation> imported;"
    beforerule = "<before> = here|keystroke;"    # to be translated, only for keystrokes extended...
    clickrule = "<click> = {click} | <modifiers> {click};"
    lastpartofgramspec = """<codesof> exported = ('Codes of') <stdkeys>+;
<count> = {count};
<repkey> = {repkey};
<norepkey> = {character}|{norepkey}|<effkey>|{punctuation};
<effkey> = (Eff|Function) {effcount};
<contextmenu> = ('context menu');
<modifiers>  = (Cap|'All-Caps'|Shift|Control|Alt)+ ;
<switchkeys> exported = "keystrokes (simple|extended)";
""" 

    gramSpec = "# placeholder of gramSpec"

    # def __init__(self, inifile_stem=None):
    #     self.language = unimacroutils.getLanguage()
    #     # here the grammar is not loaded yet, but the ini file is present
    #     self.inifile_stem = inifile_stem
    #     self.startInifile()
    #     #print 'requireTimes: %s, simpleOrExtended: %s'% (self.requireTimes, self.doKeystrokesExtended)
    #     self.constructGramSpec(inInit=1)
    #     ancestor.__init__(self)

    def constructGramSpec(self, inInit=0):
        """return the gramSpec, from gramSpecStart, with several options and tranlation steps
        """
        # first make stdkeysrule with times compulsory or voluntary:
        #if self.requireTimes:
        #    timespart = "(times)"
        #else:
        #    timespart = "[times]"
        #stdkeysrule = self.stdkeysrule % timespart
        
        if self.doKeystrokesExtended:
            startrule = "<startrule> exported = <before>+ <allkeys> | <allkeys>;"
            startSpec = ["# keystrokes extended",
                         startrule, self.stdkeysrule,
                         self.allkeysrule, self.lastkeyrule,
                         self.clickrule, self.importedrule,
                         self.beforerule]
        else:
            startrule = "<startrule> exported = <stdkeys>+;"
            startSpec = ["# keystrokes simple",
                         startrule, self.stdkeysrule]
        #print 'startSpec: %s\n\n'% startSpec
        startSpec.append(self.lastpartofgramspec)
        self.gramSpec = copy.copy(startSpec)
        # if not inInit:
        #     # this should be done only at switchkey command:
        #     self.unload()
        #     self.load(self.gramSpec)

    def initialize(self):
        #print 'self.gramspec: %s'% self.gramSpec
        self.constructGramSpec()
        self.load(self.gramSpec)
        # call for extra spoken form with difficult numbers
        # needs self.stripSpokenForm in gotResults... function!!!
        #self.addSpokenFormNumbers(self.taskCounts)
        self.countList = list(range(1, self.maxCount+1))
        self.effCountList = list(range(1, self.maxEffCount+1))  
        self.setNumbersList('count', self.countList)
        self.setNumbersList('effcount', self.effCountList)

        self.setCharactersList('character')
        self.setPunctuationList('punctuation')
        self.resetAllVars()
        self.resetVars()
        self.isIgnored = 0
        self.title = 'Unimacro grammar {self.name} (__name__) language: {language}'
        #self.activateAll()
        self.fillGrammarLists()
        self.dictateOutputState = -1 # start state for dictate <dgndictation> rule (nsformat)
            
    def gotBegin(self, moduleInfo):
        if self.checkForChanges:
            if self.checkInifile():
                self.resetAllVars()
        if self.prevModule == moduleInfo:
            return

        progInfo = unimacroutils.getProgInfo(moduleInfo)       

        self.isIgnored = 0

        if self.ignore and unimacroutils.matchWindow(self.ignore, progInfo=progInfo):
            self.isIgnored = 1
            return


        mode = self.windowPolicy(progInfo=progInfo)
        if mode == self.prevMode:
            return 
##        print 'keystrokes, new mode: ', mode
        if self.prevMode == 'inactive':
            self.activateAll()
            
        self.prevMode = mode
        ini = self.ini
        #print 'do keystrokes for mode: %s'% repr(mode)
        if mode == 'inactive':
            print(f'{self.GetName()}: deactivate: {mode}')
            self.deactivateAll()
            self.cancelMode()
            self.resetAllVars()
            return
        if mode == 'active':
            repkeySections = ['repkey']
            norepkeySections = ['norepkey']
            # print 'activate "default mode" keystrokes'
##            print '%s, mode: %s, repkeySections: %s, norepkeySections: %s'% \
##                  (self.GetName(), mode, self.repkeySections, self.norepkeySections)
        else:
            # print 'activate keystrokes with mode: %s'% repr(mode)
            # self.modeSet is set of modestrings being active:
            wantExclusive = self.modeSet & self.exclusiveModes  # both a set
            if wantExclusive:
                print(f'make keystokes mode exclusive: {wantExclusive}')
                self.setExclusive(1)
                # done at start of the grammar, possibly only do it here:
                # natlinkmain.set_on_mic_off_callback(self.on_mic_off_callback)
            #
            repkeySections = self.ini.getSectionsWithPrefix('repkey', mode)
            repkeySections.append('repkey')
            norepkeySections = self.ini.getSectionsWithPrefix('norepkey', mode)
            norepkeySections.append('norepkey')
            #print '%s, mode: %s, repkeySections: %s, norepkeySections: %s'% \
            #      (self.GetName(), mode, repkeySections, norepkeySections)

        if repkeySections != self.repkeySections:
            L = ini.get(repkeySections)
            self.setList('repkey', L)
            self.repkeySections = repkeySections[:]
            self.repkeyList = L[:]
        if norepkeySections != self.norepkeySections:
            L = ini.get(norepkeySections)
            self.setList('norepkey', L)
            self.norepkeySections = norepkeySections[:]
            self.norepkeyList = L[:]
 
    def gotResultsInit(self, words,fullResults):
        self.resetVars()
        self.buf = []
        self.codes = 0
        self.hadClick = 0
        self.fullResults = fullResults
        #print "keystrokes: %s"% words
        if fullResults and fullResults[0][1] != 'lastkey':
            self.dictateOutputState = -1  # reset, otherwise resume previous result.

    def gotResults_before(self, words, fullResults):
        """optional for here or keystroke multiple words even allowed"""
        if not self.hasCommon(words, 'here'):
            print('got words in "before" rule, but not "here": %s'% words)
            return
        # do a click, or (if next rule == contextmenu) a right click.
        button, nClick = 'left', 1
        if self.nextRule == 'contextmenu':
            button = 'right'
        #print '_keystrokes, do a "%s" mouse click'% button
        if not self.doWaitForMouseToStop():
            raise Exception("_keystrokes, mouse did not stop")
        unimacroutils.buttonClick(button, nClick)
        unimacroutils.visibleWait()
        self.hadClick = button             

    def subrule_contextmenu(self, words):
        """'context menu'"""
        if self.hadClick == 'right':
            self.hadClick = 0
            return
        self.flush()
        self.key = '{shift+f10}'
        self.flush()
        
    #def importedrule_dgnletters(self, words):
    #    print 'dgnletters: %s'% words
    #    for w in words:
    #        self.flush()
    #        posSlash = w.find('\\')
    #        if posSlash > 0:
    #            self.key = w[:posSlash]
    #        elif w[:4] in ['\\spa']:
    #            self.key = ' '
    #        elif self.hasCommon(w[1:], 'backslash'):
    #            self.key = '\\'
    #        elif self.hasCommon(w[1:],'cap'):
    #            self.cap = 1dit is een test
    #        else: # 
    #            
    def gotResults_dgndictation(self, words, fullResults):
        print(f'words of dgndictation: {words}')
        formattedOutput, self.dictateOutputState = nsformat.formatWords(words, state=self.dictateOutputState)
        self.key = formattedOutput
        self.flush()
        #
        #try:
        #    natlink.recognitionMimic(words)
        #except natlink.MimicFailed: level Hello space Has
        #    print "_keystrokes, dictate rule: cannot mimic words: %s"% words
            
                     
    def gotResults_count(self, words, fullResults):
        #print 'found count: %s'% words
        self.count = self.getNumberFromSpoken(words[0])

    # repkey gevraagd:
    def gotResults_repkey(self, words, fullResults):
        "{repkey}"
        self.flush()
        for w in words:
            self.flush()
            if self.repkeySections:
                res = self.ini.get(self.repkeySections, w)
            else:
                print(f'_keystrokes, no repkeySections for this mode: {self.prevMode}')
                return
            if res:
                self.key = res
            else:
                print(f'rep, found character or something else: {w}')
                posSlash = w.find('\\')
                if posSlash > 0:
                    self.key = w[:posSlash]
                else:
                    self.key = w

    def gotResults_effkey(self, words, fullResults):
        "( Eff|Function ) {effcount}"
        self.flush()
        count = self.getNumberFromSpoken(words[-1]) 
        self.key = '{f%s}'% count
        self.flush()

    def gotResults_norepkey(self, words, fullResults):
        "{punctuation} | {character} | {norepkey} | <effkey>"
        self.flush()
        if self.key:
            # no waiting for counts, can be done directly:
            self.flushAll()
        for w in words:
            #print 'norepkey: %s'% w
            #  Special case for context menu if the word "Here" had been said
            if self.hasCommon(w,["context menu"]) and self.hadClick:
                continue
            char = self.getCharacterFromSpoken(w)
            if char:
                self.key = char
                #print 'icalphabet: %s'% self.key
            else:
                char = self.getPunctuationFromSpoken(w)
                if char:
                    self.key = char
                else:
                    if self.repkeySections:
                        res = self.ini.get(self.norepkeySections, w)
                    else:
                        print(f'_keystrokes, no norepkeySections for this mode: {self.prevMode}')
                        return
                    res = self.ini.get(self.norepkeySections, w)
                    if not res:
                        raise ValueError(f'_keystrokes, norepkey: no code found for {w} ({self.fullResults})')
                    self.key = res
                #print 'norepkey: %s'% self.key
            self.flush()
    # 
    # kijk hier welke modifier keys (ctrl, shift, alt) gevraagd worden:
    def gotResults_modifiers(self, words, fullResults):
        """(Cap | "All-Caps" | Shift | Control | Alt)+ """
        self.flush()
        self.mod = ''
        # doorloop shift, control en alt:
        #print 'mod:', `words`
        if self.hasCommon(words, 'all-caps'):
            self.cap = max(self.cap, 2)
        elif self.hasCommon(words, 'cap'):
            self.cap = self.cap or 1
        elif self.hasCommon(words,  ['shift']):
            self.mod = 'shift+'
        # control:
        if self.hasCommon(words, ['control']):
            self.mod  = self.mod + 'ctrl+'
        # alt:
        if self.hasCommon(words, ['alt']):
            self.mod = self.mod + 'alt+'

    def gotResults_click(self, words, fullResults):
        """do click, with mouse, rule
        
        modifiers can be set
        """
        # print 'click rule: %s (modifiers: %s), prevrule: %s'% (words, self.mod, self.prevRule)
        self.flush()
        if self.key:
            self.flushAll()
        if self.prevRule is None or self.prevRule in ['startrule', 'modifiers']:
            if not self.doWaitForMouseToStop():
                print('_keystrokes, click, cancel command')
                return
        else:
            if not self.doMouseMoveStopClick():
                print("you should move the mouse a bit at least!")
                return
        possibleButtons = unimacroutils.joelsButtons
        possibleClicks = ['1', '2', '3']
        clickrules = self.getFromInifile(words[0], 'click')
        #print 'clickrules: %s'% clickrules
        if not clickrules:
            print('_keystrokes: no valid clickrule')
            return
        parts = [s.strip() for s in clickrules.split(',')]
        button = parts[0]
        if not button in possibleButtons:
            print('button should be one of %s'% possibleButtons)
            return
        if len(parts) > 1:
            
            if parts[1] not in possibleClicks:
                print(f'number of clicks ({parts[1]}) should be one of {possibleClicks}')
                return
            nClick = int(parts[1])
        else:
            nClick = 1
    
        if len(parts) > 2:
            print(f'currently only (button, clicks) allowed in clickrule, not: {clickrules}')
            return
    
        if self.nextRule == 'contextmenu' and nClick == 1:
            button = 'right'
        unimacroutils.buttonClick(button, nClick, modifiers=self.mod)
        unimacroutils.visibleWait()
        self.hadClick = button



    def doMouseMoveStopClick(self):
        """wait for mouse starting to move, stopping and then click
        """
        action("ALERT")
        if not action("WAITMOUSEMOVE"):
            action("ALERT 2")
            return None
        if not action("WAITMOUSESTOP"):
            action("ALERT 2")
            return None
        action("ALERT")
        return 1

    def gotResults_switchkeys(self, words):
        """[do] keystrokes (simple|extended)"""
        #line from fillInstanceVariables:
        #self.doKeystrokesExtended = self.ini.getBool('general', 'extended keystrokes', False)
        if self.hasCommon(words, 'simple'):
            self.doKeystrokesExtended = False
        else:
            self.doKeystrokesExtended = True
        self.ini.set('general', 'extended keystrokes', self.doKeystrokesExtended)
        self.ini.writeIfChanged()
        self.constructGramSpec() # should also reload the grammar
        #print 'after switchkeys, requireTimes: %s, simpleOrExtended: %s'% (self.requireTimes, self.doKeystrokesExtended)

    def gotResults_codesof(self, words,fullResults):
        self.codes = 1
        
    def gotResults_startrule(self, words,fullResults):
        pass # no special rule here!
    
    def gotResults_lastkey(self, words,fullResults):
        self.flush()
        self.flushAll()  
        unimacroutils.visibleWait()

    def gotResults(self, words,fullResults):
        self.flush()
        self.flushAll()
        if self.cap < 3:
            self.cap = 0
            
           
    def flush(self):
        if not self.key:
            return
        self.hadClick = 0

        if self.cap:
            self.key = unimacroutils.doCaps(self.key)

        if self.mod:
            self.key = unimacroutils.doModifier(self.key, self.mod)
        if self.count != 1:
            self.key = unimacroutils.doCount(self.key, self.count)
       
        if isinstance(self.key, list):
            print(f'_keystrokes, flush: warning, self.key is list: {self.key}')
            self.buf.extend(self.key)
        else:
            self.buf.append(self.key)
        self.resetVars()

    def flushAll(self):
        if self.buf:
            try:
                buf = ''.join(self.buf)
            except TypeError:
                print(f'---TypeError in flushAll of _keystrokes: {self.buf}')
                raise
            
            #print 'flushAll: %s'% buf  
            if self.codes:
                action(f'SCLIP {buf}')
            elif buf.find("<<") >= 0:
                action(buf)
            else:
                keystroke(buf)
            self.buf = []

    def resetVars(self):
        self.mod = ""
        self.count = 1
        self.key = ''
        if self.cap == 1:
            self.cap =0

    
    def resetAllVars(self):
        self.repkeySections = None
        self.norepkeySections = None
        self.repkeyList = self.norepkeyList = []
        self.prevModule = None
        self.prevMode = 'inactive'
        
        self.cap = 0
        self.buf = []
        self.resetVars()
       
    def gotResults_command(self, words,fullResults):
        self.cap = 0
        
        if self.hasCommon(words, ['Caps']):
            self.cap = 3
            
    def windowPolicy(self, modInfo=None, progInfo=None): 
        progInfo = progInfo or unimacroutils.getProgInfo(modInfo)
        #print 'window policy------progInfo: %s'% repr(progInfo)
        #print 'deactivaterules: %s'% self.deactivateRules
        #print 'activaterules: %s'% self.activateRules
        modeSet = []
        for mode in self.modes:
            if unimacroutils.matchWindow(self.modes[mode], progInfo=progInfo):
                modeSet.append(mode)
        self.modeSet = set(modeSet)
        if modeSet:
            return tuple(modeSet)

        if unimacroutils.matchWindow(self.activateRules, progInfo=progInfo):
            if unimacroutils.matchWindow(self.deactivateRules, progInfo=progInfo):
                return 'inactive'
            return 'active'
        return 'inactive'

       
    def showInifile(self, body=None, grammarLists=None, ini=None,
                    showGramspec=1, prefix=None, commandExplanation=None,
                    postfix=None, lineLen=60, sort=1):
        body = []
        if self.modes:
            if type(self.prevMode) in (list, tuple):
                modesString = ', '.join(self.prevMode)
            else:
                modesString = str(self.prevMode)
            body.append(f'current mode/active/inactive: {modesString}')
            if not self.prevMode:
                body.append(f'no current mode (yet): {self.prevMode}')
            elif self.prevMode == 'active':
                pass
            elif self.prevMode == 'inactive':
                # probably never comes here, because grammar is inactive at this moment
                body.append(f'inactive in {self.deactivateRules}')
            else:
                # give extra moohdes information:
                body.append('possible modes:')
                for mode in self.modes:
                    if self.modes[mode] == {mode: None}:
                        body.append(f'\t{mode}: (all windows)')
                    else:
                        body.append(f'\t{mode}: {self.modes[mode]}')
        else:
            body.append('active/inactive: {self.prevMode}')

        if self.doKeystrokesExtended:
            body.append('do keystrokes extended (with Here and mouse clicking)')
        else:
            body.append('do keystrokes simple (no mouse clicking)')
        #if self.requireTimes:
        #    body.append('require "times" when doing repeat keys')
        ancestor.showInifile(self, body=body)

    def fillDefaultInifile(self, ini):
        """initialize as a starting example the ini file obsolete

        """
        
    def fillInstanceVariables(self):
        """fills instance variables with data from inifile

        overload for grammar keystrokes

        """
        try:
            ini = self.ini
            self.maxCount = ini.getInt('general', 'max count') or 20
            self.maxEffCount = ini.getInt('general', 'max eff count') or 12
            self.doKeystrokesExtended = ini.getBool('general', 'extended keystrokes', False)
            #self.requireTimes = ini.getBool('general', 'require times', False)
            #if self.requireTimes:
            #    print 'require "times" (or translation/synonym) in keystrokes grammar when using repeatable keystrokes'

    ##        print 'fillInstantVariables for %s'% self
            self.activateRules = ini.getDict('general', 'active', {'all': None})
            self.cleanAsterixes(self.activateRules)

            self.deactivateRules = self.ini.getDict('general', 'ignore', {'none': None})
            self.cleanAsterixes(self.deactivateRules)

            modes = ini.get('modes')
            implicitModes = ini.getSectionPostfixesWithPrefix('repkey') + \
                        ini.getSectionPostfixesWithPrefix('norepkey')
            for m in implicitModes:
                if m and m not in modes:
                    #print 'implicit mode added: %s'% m
                    modes.append(m)
            if not modes:
                #print 'no modes defined for grammar: %s'% self.GetName()
                self.modes = None
            else:
                self.modes = {}
                for m in modes:
                    self.modes[m] = ini.getDict('modes', m, {m: None})
                    
            self.exclusiveModes = set(self.ini.getList('general', 'exclusive modes', []))
            for m in self.exclusiveModes:
                if m not in self.modes:
                    print(f'warning, exclusive mode "{m}" not in defined modes: {self.modes.keys()}')
            
                    
        except inivars.IniError:
            print('IniError while initialising ini variables in _keystrokes')
            print('message: %s'% sys.exc_info()[1])
            return
        
    def cleanAsterixes(self, Dict):
        """change * to None in keys and values, inplace!
        """
        if '*' in Dict:
            del Dict['*']
            Dict['all'] = None
        for prog in Dict:
            if Dict[prog] == '*':
                Dict[prog]  = None
##        print 'self.activateRules: %s'% self.activateRules
 
    def cancelMode(self):
        # ending exclusive mode if in it, also ending the mode it is in
        self.setExclusive(0)
        self.resetAllVars()
 
# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)

def unload():
    #pylint:disable=W0603, E0601
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None
 

if __name__ == "__main__":
    ## interactive use, for debugging:
    natlink.natConnect()
    try:
        thisGrammar = ThisGrammar(inifile_stem="_keystrokes")
        thisGrammar.startInifile()
        thisGrammar.initialize()
        natlinkmain.set_on_mic_off_callback(thisGrammar.cancelMode)
    finally:
        natlink.natDisconnect()
elif __name__.find('.') == -1:
    # standard startup when Dragon starts:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()
    natlinkmain.set_on_mic_off_callback(thisGrammar.cancelMode)



"""


(unimacro - natlink macro wrapper/extensions)
see http://qh.antenna.nl/unimacro/aboutunimacro.html for copyright note
    
      
"""
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroactions as actions
import natlink
import unittest
import time
import types
import UnimacroTestHelpers

#special mods, the bring up you ask for produces and other module:
specialMods = {}
specialMods['edit'] = 'uedit32'
specialMods['dragonpad'] = 'natspeak'
specialMods['voicecode'] = 'emacs'


class BringupTest(UnimacroTestHelpers.UnimacroTestHelpers):
    """Testing the Bringup in 'actions.py'

    and the corresponding shorthand commands

    the different tests can be configured.  Notice "edit" is some
    edit program that does the normal editing of .txt and .ini files
    (in case QH it is uedit32 (ultraedit)). Can be configured
    through the actions.ini configuration file.

    Also tests the do_KW (kill window) function.    

    voicecode is special, starts emacs, looks for mediator, checks the user...

    cmd is the windows command prompt    
    
    """
    first = 1
    def setUp(self):
        if self.first:
            actions.debugActions(1)
            self.first = 0
        else:
            actions.debugActions(1, openMode='a')
        self.startUser = natlink.getCurrentUser()[0]

    def tearDown(self):
        endUser = natlink.getCurrentUser()[0]
        if self.startUser != endUser:
            print('reopen user: %s'% self.startUser)
            natlink.openUser(self.startUser)
        actions.debugActions(0)
        

    def test_Should_BringUp_DragonPad(self):

        actions.UnimacroBringUp("dragonpad")
        time.sleep(2)
        self.assert_mod_partoftitle('natspeak', 'DragonPad', " when bringing up DragonPad")
        sendkeys("this should last 2 seconds")
        time.sleep(2)
        actions.do_KW()

        
    def test_Bringup_list_of_small_applications(self):
        preserveApps = []  # other apps are killed after the test
        apps = ['notepad', 'dragonpad']   ## calc renamed in Windows 10
        self.doTestBringupApplications(apps, preserveApps)

    def test_Bringup_list_of_medium_applications(self):

        preserveApps = ['pythonwin', 'edit']  # other apps are killed after the test
        apps = ['notepad', 'pythonwin', 'dragonpad', 'edit', 'cmd']
        self.doTestBringupApplications(apps, preserveApps)

    def tttest_Bringup_list_of_emacs_applications(self):

        preserveApps = ['pythonwin']  # other apps are killed after the test
        apps = ['notepad', 'emacs', 'pythonwin']
        self.doTestBringupApplications(apps, preserveApps)

    def test_Bringup_list_of_office_applications(self):

        preserveApps = []  # other apps are killed after the test
        apps = ['notepad', 'winword', 'excel']
        self.doTestBringupApplications(apps, preserveApps)
        
    def tttest_Bringup_voicecode(self):

        preserveApps = ['emacs']  #ther adepps are killed after the test
        apps = ['calc', 'emacs', 'voicecode', 'notepad']
        self.doTestBringupApplications(apps, preserveApps)
    
    def tttest_Bringup_voicecode_user_switching(self):
    
        user, dummy = natlink.getCurrentUser()
        if user == 'VoiceCode':
            mes = ['active user is: %s'% user, '',
                   'Cannot test voicecode_user_switching to VoiceCode again', '', 
                   'Please start test with another user being active.']
            actions.Message(mes)
            self.fail("start test with a user profile NOT being VoiceCode")
            

        preserveApps = ['emacs']  #ther adepps are killed after the test
        apps = ['voicecode', 'notepad']
        if not self.doTestBringupApplications(apps, preserveApps, prompt="test_Bringup_voicecode_user_switching"):
            print('test_Bringup_voicecode_user_switching skipped halfway')
            return
        newUser, dummy = natlink.getCurrentUser()
        self.assert_equal('VoiceCode', newUser, "user name should be switched to 'VoiceCode'")
        if user == newUser:
            mes = ['cannot test voicecode user coming from another user',
                   'start test from another user as "VoiceCode"']
            actions.Message(mes)

        natlink.openUser(user)
        newUser, dummy = natlink.getCurrentUser()
        self.assert_equal(user, newUser, "user name should be switched to '%s'\n"
                          "in order to be able to test the user switching part of this test"% user)
        self.doTestBringupApplications(apps, preserveApps, prompt=0)
        newUser, dummy = natlink.getCurrentUser()
        self.assert_equal('VoiceCode', newUser, "user name should be switched to 'VoiceCode', also the second time of testing")


    def doTestBringupApplications(self, apps, preserveApps, prompt=1):
        """do the testing of bringing up and destroying applications

        switching is done several times, also after killing the window.
        switching with UnimacroBringUp several times is fast and comes back
        to the same window, if not killed in between.

        In these tests the different combinations are checked.        

        """
        destroyApps = [a for a in apps if a not in preserveApps]
        if prompt:
            mes = ["Do you want to do the Bringup applications tests?", "",
                   'using %s'%apps,
                   'keeping open: %s'% preserveApps,
                   'destroying after testing: %s'% destroyApps]
            if type(prompt) == bytes:
                mes.insert(0, prompt+'\n\n')
            result = actions.YesNo(mes, "Starting the Bringup applicationsTest ")
            if not result:
                print('Bringup applications test skipped')
                return

        appString = ', '.join(apps)

        # these should correspond with "name" in [bringup app] section of actions.ini
        # call for edit actions to inspect these:
        expectedMods = dict(list(zip(apps, apps)))
        expectedMods.update(specialMods)

        # bring them up:        
        for app in apps:
            result = actions.UnimacroBringUp(app)
            if result:
                self.assert_mod(expectedMods[app], " (round 1) when bringing up %s (%s)"%
                                (app, appString))
            else:
                self.fail(" (round 1) did not get result from switching to app: %s (%s)"%
                          (app, appString))
            
        # and do it again:
        for app in apps:
            result = actions.UnimacroBringUp(app)
            if result:
                self.assert_mod(expectedMods[app], " (round 2) when bringing up %s (%s)"%
                                (app, appString))
            else:
                self.fail(" (round 2) did not get result from switching to app: %s (%s)"%
                          (app, appString))

            result = actions.UnimacroBringUp(app)
            if result:
                self.assert_mod(expectedMods[app], " (round 2, double!) when bringing up %s (%s)"%
                                (app, appString))
            else:
                self.fail(" (round 2, double!) did not get result from switching to app: %s (%s)"%
                          (app, appString))
        # do it in reverse order, kill the window, except for pythonwin and uedit32 (preserveApps):
        apps.reverse()
        for app in apps:
            result = actions.UnimacroBringUp(app)
            if result:
                self.assert_mod(expectedMods[app], " (round 3) when bringing up %s (%s)"%
                                (app, appString))
            else:
                self.fail(" (round 3, revers order) did not get result from switching to app: %s (%s)"%
                          (app, appString))
            if app not in preserveApps:
                actions.do_KW()
                self.failIf_mod(app, "(round 3) kill window did not succeed for app\n\n"
                                "(THIS MAY BE A TEST ARTEFACT OR A FAILURE OF THE KW (KillWindow) ACTION: %s (%s)"%
                                (app,appString))
                
        # again reverse order, bringup again and kill the window, except for pythonwin and uedit32:
        for app in apps:
            result = actions.UnimacroBringUp(app)
            if result:
                self.assert_mod(expectedMods[app], " (round 4, after killing previous app)) when bringing up %s (%s)"%
                                (app, appString))
            else:
                self.fail(" (round 4, after killing previous app)) did not get result from switching to app: %s (%s)"%
                          (app, appString))
            if app not in preserveApps:
                actions.do_KW()
                self.failIf_mod(app, "(round 4, after killing previous app)) kill window did not succeed for app: %s (%s)"%
                                (app,appString))
        return 1
# no main statement, run from command in _unimacrotest.py.


# _dictationproject.py

#  make other apps select and say
#  only for windows that are enabled through the "actionclasses".

#  based on work written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  October 2009-September 2010.
"""dictation implemented for apps that have the needed functions implemented in "actionclasses"

two classes are needed here, the VoiceDictation class, which intercepts the
VDct of NatSpeak,
and the CmdGrammar class, which holds global and application specific commands.

The initialisation of the VoiceDictation instance is done in the CmdGrammar, so
they have a link to each other and there is no need for a special file for the
VoiceDictation class.

When speech starts, the gotBegin of the cmdGrammar is called AND the
onTextBegin of the dictobj instance (of VoiceDictation) which is self.dictobj
inside the cmdGrammar instance.

Likewise the normal callbacks for a command are in the gotResult... functions of
cmdGrammar, and onTextChange is called when dictation has been recognised (of self.dictobj)

inside the VoiceDictation instance there is self.dictObj, which is the link to
the (intercepted) dictation object.

"""
import win32api
import win32gui
import win32con
import time
import os
import sys
import types
import re
#print 'sys.path: %s'% sys.path
import unimacro.messagefunctions as mess
try:
    # this is a little module for easy changing of test application
    # the file testermod.py (NOT included in svn) holds something like:
    ## different testings of unittestMessagefunctions:
    ##tester = "wordpad"
    ##tester = "dragonpad"
    # one of the lines should be uncommented.
    # testermod is also used by unittestMessagefunctions, which can be used
    # to test the communication with the (off focus) application.
    import testermod
    tester = testermod.tester
except ImportError:
    tester = "aligen"   # this is the window of the Kaiser Permanente application.

import windowparameters

testMode = 1
defaultWaitTime = 0.2  # change to slow down or speed up actions.

import natlink
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action

from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj

from time import sleep  

class VoiceDictation:
    dictObj = None
    ctrl = None
    # Initialization.  Create a DictObj instance 
    # activate if outside the special text window

    def initialize(self):
        if self.dictObj != None:
            D("dictObj already initialized")
            return
        self.__class__.dictObj = natlink.DictObj()
        self.__class__.app = None
        self.__class__.ctrl = None  # to be filled in getEditControl
        self.dictObj.setBeginCallback(self.onTextBegin)
        self.dictObj.setChangeCallback(self.onTextChange)
        # copy the W variables, taken from "windowparameters.py"
        W = windowparameters.PROGS[tester]
        if W:
            print('get windowsparamters from tester: %s'% tester)
            for k,v in list(W.items()):
                #print 'W: %s: %s'% (k, v)
                setattr(self, k, v)
        else:
            raise ValueError("Cannot get windowsparameters dict for tester: %s"% tester)
        self.getEditControl()
        self.inOnTextChange = 0
        # updated in updateState:
        self.lastSel = (0,0)
        self.lastSelText = ""
        self.scratchinfo = [] # lists of undo info, [(selStart, selEnd), newText, (delStart, delEnd), oldText]
        self.scratchThatCommand = 0
        self.nonFocusHndles = []  # for loosing focus
        self.hasFocus = -1  # is controlled from CmdsGrammar instance

        #self.fieldIsSelected = 0 # only set when a field is extended [x] (first only x after [x])
        
    def getAppWindow(self):
        """get application, try quick if application still exists
        
        if not, try to find (again)
        if cannot be found: try to start (if "shouldstartauto" is on)
        """
        if self.app:
            wtext = win32gui.GetWindowText(self.app)
            if wtext: return # app still active
        self.__class__.app = None # lost the appl
        appWindows = mess.findTopWindows(wantedClass=self.windowclass, wantedText=self.windowcaption)
        if len(appWindows):
            if len(appWindows) > 1:
                print('warning, more appWindows active! %s'% appWindows)
            appHndle = appWindows[0]
            self.__class__.app = appHndle
            return
        
        # not found app:
        if not self.shouldstartauto:
            return
        
        D('starting the application %s'% self.apppath)
        meHndle = win32gui.GetForegroundWindow()
        result = os.startfile(self.apppath)
        if self.testcloseapp:  # quick easy to start app probably
            sleepTime = 0.1
        else:
            sleepTime = 0.5    # maybe a slow starting app
        for iTry in range(20):
            time.sleep(sleepTime)
            appWindows = mess.findTopWindows(wantedClass=self.windowclass, wantedText=self.windowcaption)
            if appWindows: break
        else:
            print('starting %s failed, dumping apps'% self.apppath) 
            pprint.pprint(dumpTopWindows())
            return

        if meHndle:
            # try to get focus back to originating window
            try:
                unimacroutils.SetForegroundWindow(meHndle)
            except: pass
            appHndle = appWindows[0]
        self.__class__.app = appHndle
    
    def getEditControl(self):
        # get application, if not found, set ctrl to None
        self.getAppWindow()
        if not self.app:
            self.__class__.ctrl = None
            return
        currentSel = None
        if self.ctrl:
            currentSel = mess.getSelection(self.ctrl)
        if currentSel and currentSel != (0,0):
            return
        ctrls =  mess.findControls(self.app,wantedClass=self.editcontrol)
        if len(ctrls):
            editHndle = ctrls[0]
            #print 'editHndle set to: %s'% editHndle
            if len(ctrls) > 1:
                for hndle in ctrls:
                    id = win32gui.GetDlgCtrlID(hndle)
                    if id == 0xd6:
                        editHndle = hndle
                        break
                else:
                    print('did not get valid id for aligen window')
                    return
        else:
            raise ValueError("could not find the editHndle of the control: %s in application: %s"%
                          (self.editcontrol, self.apppath))
        self.__class__.ctrl = editHndle

    def wait(self, waitTime=None):
        waitTime = waitTime or defaultWaitTime
        time.sleep(waitTime)

    def saveFocus(self):
        """remember focus of current window, append to focus list
        """
        hndle = win32gui.GetForegroundWindow()
        if hndle == self.app:
            return
        if self.nonFocusHndles and self.nonFocusHndles[-1] == hndle:
            return
        title = win32gui.GetWindowText(hndle).lower()
        if (title.find("notepad") >= 0 or
            title.find("kladblok") >= 0 or
            title.find("outlook") >= 0 or
            title.find("komode") >= 0):
            print('saveWindow, ignore window of title: %s'% title)
            return
        print('adding: %s to list: %s'% (hndle, self.nonFocusHndles))
        self.nonFocusHndles.append(hndle)

    def acquireFocus(self):
        """get focus of special app, return previous handle
        
        if focus was already there, return None
        """
        if not self.app:
            raise Exception('no application for aquiring focus')
        hndle = win32gui.GetForegroundWindow()
        if hndle == self.app:
            return
        self.saveFocus()  # for next occurrence, mostly already done from gotBegin
        unimacroutils.SetForegroundWindow(self.app)
        for i in range(10):
            controlHndle = win32gui.GetForegroundWindow()
            if controlHndle == self.app: return hndle
            time.sleep(0.1)
        raise OSError('could not aquire focus for %s'% self.app)

    def looseFocus(self, hndle=None):
        """return the focus to window with hndle, or saved handle
        """
        controlHndle = win32gui.GetForegroundWindow()
        if hndle and hndle == self.app:
            raise ValueError("hndle to loose focus to IS the application handle: %s"% hndle)
        
        if controlHndle == self.app:
            if hndle and hellonot (self.nonFocusHndles and self.nonFocusHndles[-1] == hndle):
                self.nonFocusHndles.append(hndle)
        else:
            if hndle:
                if hndle == controlHndle:
                    return  # OK
                if not(self.nonFocusHndles and self.nonFocusHndles[-1] == hndle):
                    self.nonFocusHndles.append(hndle)
                    
        print('loosing focus, hndle: %s'% self.app)     
        # now do the work:
        if not self.nonFocusHndles:
            print('cannot loose focus, no stack')
            #if controlHndle == self.app:
            #    action("SSK {alt+tab}")
            #    self.wait(0.3)
            return
        while self.nonFocusHndles:
            hndle = self.nonFocusHndles.pop()
            try:
                print('loosing focus, to:%s (rest: %s)'% (hndle, self.nonFocusHndles))          
                unimacroutils.SetForegroundWindow(hndle)
                self.wait()
                self.nonFocusHndles.append(hndle)
                return
            except:
                print('exception loosing focus to %s, try previous on list'% hndle)
                continue
            raise Exception('could not loose focus an favour of %s'% hndle)
        else:
            if controlHndle == self.app:
                print('cannot loose focus, stack exhausted')
                #action("SSK {alt+tab}")
                #self.wait(0.3)
            
            
    def activateMenuItem(self, menuItem):
        """do menu time through mess
        """
        if not self.app:
            print('activeMenuItem %s, no control active: %s'% (menuItem,self.app))
        menuToDo = getattr(self, menuItem, None)
        if menuToDo:
            mess.activateMenuItem(self.app, menuToDo)
            self.wait()
        else:
            print('no valid menu item in this application: %s'% menuItem)
            
    def sendKey(self, keyOrKeys):
        """send key(s) through messagefunctions to control
        """
        if not self.ctrl:
            print('sendKey %s, no control active: %s'% (menuItem,self.ctrl))
        if keyOrKeys:
            mess.sendKey(self.ctrl, keyOrKeys)
            self.wait()
        else:
            print('sendKey, no keys specified')
    
    def sendEnterKey(self, numTimes=1):
        t = self.aftertext*numTimes
        t = self.correctForNewlines(t)
        delStart, delEnd = self.getSelection()
        selStart = selEnd = delStart + len(t)
        self.onTextChange(delStart,delEnd,t,selStart,selEnd)
        self.updateState()
        
    def onTextBegin(self,moduleInfo):
        """at start of each utterance,
        
        The control of the started application is done from the CmdsGrammar
        in each gotBegin callback.
        """
        pass  # gets controlled from CmdsGrammar.gotBegin!
        #if not self.app:
        #    self.getEditControl()
        #if not self.app:
        #    return
        ##D('win dict, text begin')
        #self.updateState()
        
    #def selectField(self):
    #    """if field ([x] is selected), do all three of the characters
    #    """
    #    self.fieldIsSelected = 0
    #    dct = self.dictObj
    #    sel = dct.getTextSel()
    #    if sel[0] == 0: return
    #    t = dct.getText(sel[0], sel[1])
    #    print 'selection: %s, %s, selected text: |%s|'% (sel[0], sel[1], t)
    #    if not t:
    #        print 'return from select field, result: %s'% self.fieldIsSelected
    #        return
    #    if len(t) > 1 and t[0] == '[' and t[-1] == ']':
    #        self.fieldIsSelected = 1
    #        print 'return from select field (got a field), result: %s'% self.fieldIsSelected
    #        return
    #    print 'selected now: |%s|'% t
    #    tBefore = dct.getText(sel[0]-1, sel[0])
    #    tAfter = dct.getText(sel[1], sel[1]+1)
    #    if tBefore[0] == '[' and tAfter[-1] == ']':
    #        newSel = (sel[0]-1, sel[1]+1)
    #        dct.setTextSel(newSel[0], newSel[1])
    #        self.setSelection(newSel[0], newSel[1])
    #        time.sleep(0.3)
    #        self.looseFocus()
    #        self.fieldIsSelected = 1
    #    print 'end of select field, result: %s'% self.fieldIsSelected
    # We get this callback when something in the dictation object changes
    # like text is added or something is selected by voice.  We then update
    # the edit control to match the dictation object.

    def onTextChange(self,delStart,delEnd,newText,selStart,selEnd):
        if self.inOnTextChange:
            return
        self.inOnTextChange = 1
        D('onTextChange: %s, %s, |%s|, %s, %s'% ( delStart, delEnd, repr(newText), selStart, selEnd) )
        
        if not self.ctrl:
            D("do not have ctrl control")
            return
        dct = self.dictObj
        dct.setLock(1)
        getFocus = (selEnd - selStart > 1) # at start of call
        if newText:
            if self.lastSel == (delStart, delEnd):    
                dct.setText(self.lastSelText, delStart, selEnd)  # empty now
            else:
                print('window changed, update again')
                self.updateState()
            self.insertText(delStart, delEnd, newText)


        else:
            # loosing scratch that info:
            self.scratchinfo = []
            if delStart < delEnd:
                print('selection to delete: %s, %s'% (delStart, delEnd))
            if selStart < selEnd:
                self.setSelection(selStart,selEnd) # only set the window...
            else:
                self.insertText(selStart, selEnd, newText)
                
                #selStart, selEnd, newText = self.adjustAtEndOfText(selStart, selEnd, "")
                #if newText:
                #else:
                #    self.setSelection(selStart,selEnd) # only set the window...
                    
        dct.setLock(0)
        self.inOnTextChange = 0
        if getFocus:
            unimacroutils.SetForegroundWindow(self.app)
        

    def insertText(self, delStart, delEnd, newText):
        # adjust at start:
        dct = self.dictObj
        newText = newText.replace('\n', '')                         
        delStart, newText = self.adjustAtStartOfText(delStart, newText)
        # make visible cursor of 1 space at the end
        selStart = delStart + len(newText)
        selStart, selEnd, newText = self.adjustAtEndOfText(selStart, delEnd, newText)
        # now go in internal buffer:
        dct.setText(newText, delStart, delEnd)
        dct.setTextSel(selStart, selEnd)

        # now the window buffer (through mess functions):                
        D("setting: |%s| to pos %s, %s"% (newText, delStart, delEnd))
        self.setSelection(delStart, delEnd)
        self.replaceSelection(newText)
        D("setting selection to pos %s, %s"% (selStart, selEnd))
        self.setSelection(selStart,selEnd)
        #if getFocus:
        #    unimacroutils.SetForegroundWindow(self.app)
        if not self.scratchThatCommand:
            self.scratchinfo.append( [(selStart, selEnd), newText,
                                  (delStart, delEnd), self.lastSelText])
        
    def scratchThat(self):
        if self.lastSel:
            selStart, selEnd = self.lastSel
            if selStart < selEnd and self.lastSelText != " ":
                self.scratchinfo = []
                self.insertText(selStart, selEnd, "")                
                return
                
        if self.scratchinfo:
            info = self.scratchinfo.pop()
        else:
            print('scratch that buffer empty')
            return            
        (selStart, selEnd), newText, (delStart, delEnd), prevText = info
                              
        D('scratch that: selStart, selEnd: %s, %s, latest text: %s'% (selStart, selEnd, newText))
        D('to replace with delStart, delEnd: %s, %s, previous text: %s'% (delStart, delEnd, prevText))
        myDelStart = selEnd - len(newText)
        myDelEnd = selEnd
        myNewText = prevText
        mySelEnd = myDelStart + len(prevText)
        mySelStart = mySelEnd - selEnd + selStart
        dct = self.dictObj
        dct.setLock(1)
        dct.setText(myNewText, myDelStart, myDelEnd)
        dct.setTextSel(mySelStart, mySelEnd)
        self.setSelection(myDelStart, myDelEnd)
        self.replaceSelection(myNewText)
        self.setSelection(mySelStart, mySelEnd)
        dct.setLock(0)
            
    #def undoDctBuffer(self, delStart, delEnd, newText, selStart, selEnd):
    #    """go back to previous situation in internal buffer
    #    """
    #    dct = self.dictobj
    #    dct.setText("", delStart, selEnd)
        
    def adjustAtStartOfText(self, delStart,newText):
        """insert space at start, possibly change delStart,
        
        return delStart and the new newText
        """
        dct = self.dictObj
        if self.lastSelText == '[x]':
            print('field found1')
            newText = self.capitalizeNoSpace(newText)
            return delStart, newText
        elif self.lastSelText.lstrip() == '[x]':
            print('field found2')
            newText = ' ' + self.capitalizeNoSpace(newText)
            return delStart, newText
        
        textBefore = dct.getText(max(0, delStart-20), delStart)
        textBefore = textBefore.split('\n')[-1]
        textBeforeStripped = textBefore.rstrip(' ')
        nSpacesBefore = len(textBefore) - len(textBeforeStripped)
        if nSpacesBefore:
            D("spaces before: %s, text before stripped: |%s|"%
                         (nSpacesBefore, repr(textBeforeStripped[-2:])))
        elif textBefore:
            D("text before: |%s|"% textBefore[-2:])
        else:
            D("no relevant text before cursor")

        if not textBeforeStripped:
            # start of buffer of start of line:
            newText = self.capitalizeNoSpace(newText)
            return delStart, newText

        # first sort out capitalization:
        newText = newText.lstrip(' ')
        if textBeforeStripped[-1] == ".":
            if textBefore[-2] in string.letters:
                # capitalize next
                newText = self.capitalizeNoSpace(newText)
        elif textBeforeStripped[-1] in "!?":
            newText = self.capitalizeNoSpace(newText)
        # next sort out spacing:
        if nSpacesBefore == 0:
            if textBefore[-1] in string.letters:
                if newText and newText[0] in string.letters:
                    newText = " " + newText
            elif textBefore[-1] in string.punctuation:
                if newText and newText[0] in string.letters:
                    newText = " " + newText
        elif newText and newText[0] in string.punctuation:
            delStart -= nSpacesBefore
        return delStart, newText
    
    def adjustAtEndOfText(self, selStart, delEnd, newText):
        """adjust at end of buffer
        remove spaces and select one extra space for visibility

        return selStart, selEnd, newText
        """
        dct = self.dictObj
        textAfter = dct.getText(delEnd, delEnd+20)

        newTextStripped = newText.rstrip(' ')
        nSpaces = len(newText) - len(newTextStripped)
        if nSpaces:
            selEnd = selStart
            selStart = selStart - nSpaces
            return selStart, selEnd, newText
        else:
            if not (textAfter and textAfter[0] == " "):
                newText = newText + " "
            selEnd = selStart + 1
            return selStart, selEnd, newText

    def capitalizeNoSpace(self, newText):
        """remove spaces at start, capitalize first character
        """
        D("capitalize no space before |%s|"% newText)
        while newText and newText[0] == ' ':
            newText = newText[1:]
        if newText:
            newText = newText[0].upper() + newText[1:]
        D("capitalize no space result |%s|"% newText)
        return newText


    def getInfoBeforeCursor(self, bufferBefore):
        """return the last non space character  and the number of spaces at the end
        
        if at beginning of line or buffer, (None, numSpaces) is returned
        """
        bufferStripped = bufferBefore.rstrip(' ')
        nSpaces = len(bufferBefore) - len(bufferStripped)
        if bufferStripped:
            last = bufferStripped[-1]
            if last == '\n':
                last = None
        else:
            last = None
        return last, nSpaces
    
    def capitaliseFirstCharacter(self, text):
        """leave leading spaces intact, cap first word
        """
        if not text.lstrip(' '):
            return text
        tStripped = text.lstrip(' ')
        if tStripped == text:
            return text[0].upper() + text[1:]
        nSpaces = len(text) - len(tStripped)
        return ' '*nSpaces + tStripped[0].upper() + tStripped[1:]
        
    def correctForNewlines(self, t):
        """change \n into \r\n or \r\n into \n
        depending on self.linesep
        """
        if t.find('\n') == -1:
            return t
        
        if self.linesep == '\n':
            if t.find('\r') >= 0:
                t = t.replace('\r', '')
        elif self.linesep == '\r\n':
            t = t.replace('\n', '\r\n')
            t = t.replace('\r\r\n', '\r\n')
        print('correctedForNewlines: %s'% repr(t))
        return t

    def updateState(self, checkField=1):
        dct = self.dictObj
        if testMode:
            beforeLength = dct.getLength()
            beforeText = dct.getText(0, beforeLength)
        if self.ctrl:
            text = self.getWindowText()
            #text = text.replace('\r', '')   # only \n for internal buffer
            selStart,selEnd = self.getSelection()
            #D("updateState, ctrl: selStart, selEnd: %s, %s"% (selStart, selEnd))
        visStart,visEnd = self.getVisibleRegion()
        if (visStart, visEnd) == (None, None):
            visStart, visEnd = 0, len(text)
        dct.setLock(1)
        dct.setText(text,0,0x7FFFFFFF)
        dct.setTextSel(selStart,selEnd)
        dct.setVisibleText(visStart, visEnd)
        self.lastSel = (selStart, selEnd)
        self.lastSelText = text[selStart:selEnd]
        D("dictObj set with text: %s (visible: %s, %s)"% (len(text), visStart, visEnd))
        D("part of text (10 left from selStart): |%s|"% repr(dct.getText(min(0, selStart-10), 0x7FFFFFFF)))
        if 0:  # testMode:
            afterLength = dct.getLength()
            afterText = dct.getText(0, afterLength)
            if afterLength != beforeLength:
                D("length after updating not the same: before: %s, after: %s"%
                  (beforeLength, afterLength))
            if afterText != beforeText:
                D("textafter updating not the same: before:\n|%s|\nafter:\n|%s|\n"%
                  (repr(beforeText), repr(afterText)))

        if selStart > 0 and self.lastSelText == 'x':
            if text[selStart-1:selStart] == '[' and text[selEnd:selEnd+1] == ']':
                print('extend selection')
                selStart -= 1
                selEnd += 1
                self.setSelection(selStart, selEnd)
                dct.setTextSel(selStart, selEnd)
                self.lastSel = (selStart, selEnd)
                self.lastSelText = text[selStart:selEnd]
        if self.lastSelText.strip() == '[x]':
            print('lastSelText, want to loose focus!!: |%s|'% self.lastSelText)
            
            hndle = win32gui.GetForegroundWindow()
            newHasFocus = (hndle == self.app)
            if newHasFocus != self.hasFocus:
                print('focus changed during updateState/gotBegin, now: %s'% newHasFocus)
                self.hasFocus = newHasFocus
            if newHasFocus:
                print('loosing focus')
                self.looseFocus()
                hndle = win32gui.GetForegroundWindow()
                if hndle == self.app:
                    print('could not loose focus!!')
        dct.setLock(0)
            
       
    def clearBoth(self):
        """for testing, clear dictobj and current window
        """
        if not self.ctrl:
            print("no current control active...")
        mess.setEditText(self.ctrl, "")
        self.updateState()
        

    def getSelection(self):
        """get the selection of the edit control
        
        return a 2 tuple
        """
        return mess.getSelection(self.ctrl)
    def setSelection(self, start, end):
        """change the selection of the edit control
        """
        mess.setSelection(self.ctrl, start, end)

    def getWindowText(self):
        """get the buffer from the edit control
        
        should go smarter in
        """
        #print 'getting buffer'
        tList = mess.getEditText(self.ctrl)
        #print '\n\ntList one:\n|%s|\n'% repr("||".join(tList))
        ## remove internal (for scrolling purposes) newline characters:
        #tList = [t.replace('\n', '') for t in tList]
        #print 'tList two:\n|%s|\n\n'% repr("||".join(tList))
        #tList = [t.replace('\n', '') for t in tList]
        #print 'tList: '
        #for i, t in enumerate(tList):
        #    print '%s: %s'% (i, repr(t))
        # the \r at end of lines remain and are the real paragraph signs
        # the intermediate \n characters are ignored in the dictobj.
        # apparently only visible for copy paste, not relevant for cursor count.
        return ''.join(tList)
        
    def replaceSelection(self, output):
        """overwrite selection with output
        """
        mess.replaceEditText(self.ctrl, output)
 
        
    def getVisibleRegion(self):
        """Utility subroutine which calculates the visible region of the edit

         control and returns the start and end of the current visible region.
        """
        return None, None
        if self.ctrl:
            buf = self.getWindowText()
            return 0, len(buf)
            
            top,bottom,left,right = self.ctrl.GetClientRect()
            firstLine = self.ctrl.GetFirstVisibleLine()
            visStart = self.ctrl.LineIndex(firstLine)

            lineCount = self.ctrl.GetLineCount()
            lastLine = lineCount
            for line in range(firstLine+1,lineCount):
                charInLine = self.ctrl.LineIndex(line)
                left,top = self.ctrl.GetCharPos(charInLine)
                if top >= bottom:
                    break
                lastLine = line

            visEnd = self.ctrl.LineIndex(lastLine+1)
            if visEnd == -1:
                visEnd = len(self.ctrl.GetWindowText())
            return visStart,visEnd
        else:
            return None, None
    
    # Call this function to cleanup.  We have to reset the callback
    # functions or the object will not be freed.
        
    def terminate(self):
        if self.dictObj:
            self.dictObj.deactivate()
            self.dictObj.setBeginCallback(None)
            self.dictObj.setChangeCallback(None)
            self.__class__.dictObj = None
            self.__class__.ctrl = None
            self.__class__.app = None

ancestor = natbj.IniGrammar
class CmdsGrammar(ancestor):
    name = "global dictation"
    alwaysrules = ['insertfragments', 'nextfield', 'selectclearall', 'testrule', 'getloosefocus']
    outsidefocusrules = alwaysrules + ['gotobeginend', 'scratch', 'deletecharswords']
    
    gramSpec = """
<insertfragments> exported = (insert) {fragments};
<gotobeginend> exported = goto (begin|end|top|bottom);
<nextfield> exported = jump | next field;
<selectclearall> exported = (clear|select) all;
<scratch> exported = scratch [that] [{scratchcounts}];
                     
<deletecharswords> exported = (backspace | delete) {backspacedeletecounts};
<getloosefocus> exported = (get|acquire|loose) focus;
<testrule> exported = global dictation test ({global_tests}|all);
"""
    def __init__(self):
        ancestor.__init__(self)
        self.dictobj = None

    def gotBegin(self, modInfo):
        hndle = modInfo[2]
        dct = self.dictobj
        dct.getEditControl() # also notices when app has been killed.
        
        if self.prevHndle == hndle:
            if dct.app and not dct.hasFocus:
                dct.updateState()
            return   # no change
        if modInfo[1] == "Spell":
            #print "Spell window"
            return
        self.prevHndle = hndle
        dct.saveFocus() # loosing the focus...
        # focus changed, so hasFocus state can be changed:
        if not dct.app:
            print('no dictation application ready: %s'% dct.app)
            dct.dictObj.deactivate()
            self.activateSet(self.outsidefocusrules)
            return
        dct.updateState()
        hndle = win32gui.GetForegroundWindow()
        if dct.app == hndle:
            if dct.hasFocus != 1:
                self.activateSet(self.alwaysrules, exclusive=0)
                dct.hasFocus = 1
                print('got focus')
                dct.dictObj.deactivate()
        else:
            if dct.hasFocus != 0:
                # lost focus now:
                print("loosing focus")
                self.activateSet(self.outsidefocusrules)
                dct.hasFocus = 0
                dct.dictObj.activate(0)
  
    def wait(self, waitTime=None):
        """some short time after each command
        
        waits after actions with messagefunctions are in the wrapper functions
        of the VoiceDictation instance
        """
        self.dictobj.wait(waitTime)
        
    def initialize(self):
        self.load(self.gramSpec)
        #print 'cmdGrammar loaded'
        # if switching on fillInstanceVariables also fill number1to9 and number1to99!
        self.switchOnOrOff()

        self.prevHndle = 0
        if not self.dictobj:
            self.dictobj = VoiceDictation()
            self.dictobj.initialize()
        
    def gotResults_insertfragments(self,words,fullResults):
        """special commands"""
        dct = self.dictobj
        #print 'includefragments: %s'% `words`
        for w in words[1:]:
            t = self.getFromInifile(w, 'fragments')
            if w.lower() == "impression":
                self.gotResults_gotobeginend(["goto", "top"], fullResults)
                self.wait()
                dct.sendKey("{end}")
            if w.lower() in ["findings", "impression", "history", "technique", "comparison"]:
                dct.sendEnterKey(2)
                self.wait()
            if t:
                t = dct.correctForNewlines(t)
                delStart, delEnd = dct.getSelection()
                selStart = selEnd = delStart + len(t)
                dct.onTextChange(delStart,delEnd,t,selStart,selEnd)
                dct.updateState()
                
            if w.lower() in ["findings", "impression", "history", "technique", "comparison"]:
                dct.sendKey(" ")
                self.wait()
                dct.updateState()             
                 
    def gotResults_gotobeginend(self,words,fullResults):
        """special commands"""
        dct = self.dictobj
        w = words[-1]
        if w in ("bottom", "end"):
            dct.setSelection(-1,-1)
        elif w in ("top", "begin"):
            dct.setSelection(0,0)
        dct.updateState()

    def gotResults_selectclearall(self,words,fullResults):
        """try to select (and clear) all text through messages menu commands
        
        focus does not have to change
        """
        dct = self.dictobj
        if dct.app:
            #prevFocus = dct.acquireFocus()
            #print 'prevFocus %s , newFocus: %s'% (prevFocus, dct.app)
            #self.wait(1)
            dct.activateMenuItem("commandselectall")
            self.wait()
            if self.hasCommon(words, "clear"):
                print("clear window")
                #keystroke("{del}")
                dct.sendKey("{del}")
            #self.wait(2)
            #dct.looseFocus(prevFocus)
        else:
            D("Not a valid menu command in app %s: %s"% (dct.app, dct.commandselectall))

    def gotResults_nextfield(self,words,fullResults):
        """special commands"""
        dct = self.dictobj
        if dct.app and dct.commandnextfield:
            prevFocus = dct.acquireFocus()
            dct.activateMenuItem("commandnextfield")
            #this is a test.dct.updateState()
            dct.looseFocus(prevFocus)
            hndle = win32gui.GetForegroundWindow()
            if hndle == dct.app:
                print('in target application')
                self.wait(1)
                dct.sendKey("{backspace}")
                self.wait()
                natlink.recognitionMimic(["\\Cap"])
            #dct.selectField()
        else:
            D("Not a valid menu command in app %s: %s"% (dct.app, dct.commandnextfield))

    def gotResults_scratch(self,words,fullResults):
        """undo one or more steps of the undo list"""
        nScratch = 1
        t = self.getFromInifile(words[-1], 'scratchcounts')
        if t:
            nScratch = int(t)
            print('scratch that %s (%s)'% (nScratch, t))
        dct = self.dictobj
        if dct.app:
            for i in range(nScratch):
                dct.scratchThat()
        else:
            D("scratch that invalid, app not active: %s"% dct.app)
              
    def gotResults_deletecharswords(self,words,fullResults):
        """delete one of more characters (later words) 
        to the right or to the left.
        """
        nToDo = 1
        t = self.getFromInifile(words[-1], 'backspacedeletecounts')
        if t:
            nToDo = int(t)
        
        if self.hasCommon(words, 'delete'):
            Key = "{delete}"
        elif self.hasCommon(words, 'backspace'):
            Key = "{backspace}"
        else:
            raise ValueError("invalid keyword for rule deletecharswords: %s"% words[0])
            
        dct = self.dictobj
        if dct.app:
            for i in range(nToDo):
                dct.sendKey(Key)
        else:
            D("deletecharswords invalid, app not active: %s"% dct.app)
              
    def gotResults_getloosefocus(self,words,fullResults):        
        """get or loose focus of the special dictation window
        """
        dct = self.dictobj
        if self.hasCommon(words, 'loose'):
            dct.looseFocus()
        else:
            dct.acquireFocus()


    def gotResults_testrule(self,words,fullResults):        
        """special for testing the synchronisation between dictObj and real window
        
        one: just dictate a few words
        two: dictate on two lines (with empty in between)
        three: select the word "second" on the second line of text
        four:  put a larger text in the window (so it scrolls) and select text.
        
        
        """
        dct = self.dictobj
        dct.clearBoth()
        test = words[-1]
        if test == 'all':
            # get list of keywords in the inifile:
            tests = self.ini.get('global_tests')
            if tests:
                print('testwords: %s'% tests)
                for lastWord in tests:
                    natlink.recognitionMimic(["global", "dictation", "test", lastWord])
                return
            else:
                print('found no "global dictations tests"')
                return 
                
        if test == "one":
            # test just one line of text, to be found in dictobj and in actual window (through messagefunctions)
            natlink.recognitionMimic(["hello", "test", "one"])
            time.sleep(0.5)
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            messText = dct.getWindowText()
            dct.updateState()
            afterLength = dct.dictObj.getLength()
            afterText = dct.dictObj.getText(0, beforeLength)
            expected = ["Hello test one " + dct.aftertext,
                        "Hello test one" + dct.aftertext] # 0: other window, 1: window has focus
            if not self.assert_equal_strings(expected, beforeText, "test %s; beforeText (dctobj) matches expected??"% test):
                return
            if not self.assert_equal_strings(beforeText, messText, "test %s; beforeText, messText equal test"% test):
                return
            if not self.assert_equal_strings(beforeText, afterText, "test %s; beforeText, afterText equal test"% test):
                return
            print('test %s OK--'% test)
            return 1  # OK
        elif test in ('two', 'three'):
            thirdWord = "second"
            if test == 'three': thirdWord = "third"
            # send a three line (two paragraph) text, and test the result:
            natlink.recognitionMimic(["hello", "\\New-Paragraph", thirdWord, "test"])
            time.sleep(0.5)
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            messText = dct.getWindowText()
            dct.updateState()
            if test == ('two', 'three'):
                afterLength = dct.dictObj.getLength()
                afterText = dct.dictObj.getText(0, beforeLength)
                expected = ["Hello\r\%s test %s" (thirdWord.capitalize(), dct.aftertext),
                            "Hello\r\r%s test"% (thirdWord.capitalize(),dct.aftertext)]  # 0: other window, 1: window has focus
                if not self.assert_equal_strings(expected, beforeText, "test %s; beforeText (dctobj) matches expected??"% test):
                    return
                if not self.assert_equal_strings(beforeText, messText, "test %s; beforeText, messText equal test"% test):
                    return
                if not self.assert_equal_strings(beforeText, afterText, "test %s; beforeText, afterText equal test"% test):
                    return
                if test == 'two':
                    print('test %s OK--'% test)
                    return 1
            if test == 'three':
                # select a word on the third line (second paragraph)
                natlink.recognitionMimic(["select", "third"])
                time.sleep(0.5)
                beforeSel = dct.dictObj.getTextSel()
                dctSelection = beforeText[beforeSel[0]:beforeSel[1]]
                messSel = mess.getSelection(dct.ctrl)
                messSelection = messText[messSel[0]:messSel[1]]
                expected = "Third "
                if not self.assert_equal_strings(expected, dctSelection, "test %s; selection from dctobj "% test):
                    return
                if not self.assert_equal_strings(expected, messSelection, "test %s; selection from message window "% test):
                    return
                print('test %s OK--'% test)
                return 1 # OK
        if test == "four":
            natlink.recognitionMimic(["test", "four", "\\New-Paragraph", "there", "we", "go"])
            natlink.recognitionMimic(["insert", "test", "fracture"])
            natlink.recognitionMimic(["hello", "after", "fracture"])
            natlink.recognitionMimic(["select", "hello", "after"])
            time.sleep(0.5)
            
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            beforeSel = dct.dictObj.getTextSel()
            dctSelection = beforeText[beforeSel[0]:beforeSel[1]]
            messText = dct.getWindowText()
            messSel = mess.getSelection(dct.ctrl)
            messSelection = messText[messSel[0]:messSel[1]]
            expected = "Hello after "
            if not self.assert_equal_strings(expected, dctSelection, "test %s; selection from dctobj "% test):
                return
            if not self.assert_equal_strings(expected, messSelection, "test %s; selection from message window "% test):
                return
            print('test %s OK--'% test)
            return 1 # OK
        if test == 'five':
            natlink.recognitionMimic(["test", ",\\comma", "five"])
            natlink.recognitionMimic(["select", "test"])
            natlink.recognitionMimic(["hello", "again", "five"])
            self.wait()
            
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            beforeSel = dct.dictObj.getTextSel()
            messText = dct.getWindowText()
            messSel = mess.getSelection(dct.ctrl)
            expected = "Hello again, five"
            if not self.assert_equal_strings(expected, beforeText, "test %s; text from dctobj "% test):
                return
            if not self.assert_equal_strings(expected, messText, "test %s; text from window "% test):
                return
            print('test %s OK--'% test)
            return 1 # OK
        if test == 'six':
            # try to make a field, and see if next dictate removes the field and caps the first char
            natlink.recognitionMimic(["insert", "impression"])
            dct.sendKey(" ")
            dct.sendKey("[")
            dct.sendKey("x")
            dct.sendKey("]")
            natlink.recognitionMimic(["select", "x\\xray"])
            self.wait()
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            beforeSel = dct.dictObj.getTextSel()
            messText = dct.getWindowText()
            messSel = mess.getSelection(dct.ctrl)
            expected =   '\r\rImpression: [x]\r'
            if not self.assert_equal_strings(expected, beforeText, "test %s; text from dctobj "% test):
                return
            if not self.assert_equal_strings(expected, messText, "test %s; text from window "% test):
                return
            dctSelection = beforeText[beforeSel[0]:beforeSel[1]]
            messSelection = messText[messSel[0]:messSel[1]]
            expected = "[x]"
            if not self.assert_equal_strings(expected, dctSelection, "test %s; selection from dctobj (selected field) "% test):
                return
            if not self.assert_equal_strings(expected, messSelection, "test %s; selection from message window (selected field) "% test):
                return
            
            print('test %s OK--'% test)
            return 1 # OK
            
        if test == 'seven':
            # try to make a field, and see if next dictate removes the field and caps the first char
            natlink.recognitionMimic(["insert", "findings"])
            dct.sendKey(" ")
            dct.sendKey("[")
            dct.sendKey("x")
            dct.sendKey("]")
            natlink.recognitionMimic(["select", "x\\xray"])
            self.wait()
            natlink.recognitionMimic(["insert", "test", "bone"])
            beforeLength = dct.dictObj.getLength()
            beforeText = dct.dictObj.getText(0, beforeLength)
            beforeSel = dct.dictObj.getTextSel()
            messText = dct.getWindowText()
            messSel = mess.getSelection(dct.ctrl)
            expected =  '\r\rImpression:\r\r1. Text in field \r'
            if not self.assert_equal_strings(expected, beforeText, "test %s; text from dctobj "% test):
                return
            if not self.assert_equal_strings(expected, messText, "test %s; text from window "% test):
                return
            print('test %s OK--'% test)
            return 1 # OK
            
            
        
            

    def assert_equal_strings(self, one, two, message):
        """one is expected string or list of strings
        
        if one is a list, 0: focus on, 1: focus elsewhere
        returns 1 if OK, 0 if error.
        """
        dct = self.dictobj
        hndle = win32gui.GetForegroundWindow()
        
        if hndle == dct.app:    
            message = 'testing equal strings, focus ON\n-- %s --'% message
            if type(one) == list:
                one = one[1] # second alternative on list
        else:
            message = 'testing equal strings, focus OFF (extra space often)' + message
            if type(one) == list:
                one = one[0]  # first alternative on list
        if one == two:
            return 1
        print('\nerror at test'+ message) 
        
        if len(one) != len(two):
            print('unequal lengths: %s, %s'% (len(one), len(two)))
        for i, (charOne, charTwo) in enumerate(zip(one, two)):
            if charOne != charTwo:
                print("strings different on pos: %s"% i)
                print("one: %s, two: %s"% (repr(charOne), repr(charTwo)))
                print("command start of string: %s"% repr(one[:i]))
                print("one: %s"% repr(one))
                print("two: %s"% repr(two))
                break
        else:
            # identical start, so tail must be different.
            if len(one) > len(two):
                print('one is longer with: |%s|'% repr(one[len(two):]))
            else:
                print('two is longer with: %s'% repr(two[len(one):]))

    def assert_equal_selection(self, one, two, message):
        print('testing equal selection: %s'% message)
        if len(one) != len(two):
            print('unequal lengths: %s, %s'% (len(one), len(two)))
        if one != two:
            print('unequal selections: %s, %s'% (repr(one), repr(two)))
            return
        print('test OK----------------')
        return 1
        
# debug print, only if testMode is o
def D(t):
    if testMode:
        print(t)

# standard stuff Joel:
cmdsGrammar = CmdsGrammar()
if cmdsGrammar.gramSpec:
    cmdsGrammar.initialize()
else:
    cmdsGrammar = None

#dictObj = VoiceDictation() # now called from commands grammar... may have to change this
#dictObj.initalize()

#dictObj = None
def unload():
    global cmdsGrammar, dictObj
    if cmdsGrammar:
        cmdsGrammar.dictobj.terminate()
        cmdsGrammar.unload()
    cmdsGrammar = None
    #if dictObj:
    #    dictObj.terminate()
    #    dictObj = None
        
messWindow = None
messTitle = "Messages from Python Macros"
def changeCallback(type,args):
    global messWindow
    # not active without special version of natlinkmain,
    # call the cancelMode, to switch off exclusive mode when mic toggles:
    # try to find the messages window at each mic on and minimize if found:
    if type == 'mic' and args == 'on':
        if testMode: return # only in production mode...
        #print 'mic on, messWindow: %s'% messWindow
        if messWindow:
            try:
                title = win32gui.GetWindowText(messWindow)
                if title != messTitle:
                    print('title does not match: actual: |%s|, wanted: |%s|'% (title, messTitle))
                    messWindow = None
            except:
                messWindow = None
        if not messWindow:
            messWindows = mess.findTopWindows(wantedText="Messages from Python Macros")
            if messWindows:
                messWindow = messWindows[0]
        if not messWindow:
            print('didnt find messages window')
            return
        #toMin = win32con.SW_MINIMIZE
        toMin = win32con.SW_SHOWMINIMIZED
        info = list( win32gui.GetWindowPlacement(messWindow) )
        #print 'info1: %s, SW_SHOWMINIMIZED: %s'% (info[1], toMin)
        info[1] = toMin
        win32gui.SetWindowPlacement(messWindow, tuple(info) )
            
        
